""" Module with classes that wrap commonly used utility functions from pytorch
"""

import torch
from torch import nn

class Concat:
    """ Simple class that wraps torch.cat
    
    Attributes
    ----------
    dim : int
        Which dimension to concatenate
    """

    def __init__(self, dim):
        self.dim = dim

    def __call__(self, *input):
        return torch.cat(input, dim=self.dim)

class SimpleSelect:
    """ Simple selection class that wraps the selection of the first
    dimension
    
    Attributes
    ----------
    sel : int
        Which element to select
    """

    def __init__(self, selection_axis):
        self.sel = selection_axis
        
    def __call__(self, X):
        return X[self.sel]

class SliceSelect:
    """ Selection class that wraps the indexing of Sequentials via slices
    
    Attributes
    ----------
    slice : slice
        Slice to index input
    """

    def __init__(self, sel_slice):
        self.slice = sel_slice
        
    def __call__(self, X):
        return X[self.slice]



"""Module with functions to construct a holistic torch.nn.Model from a set of
models."""
import logging

import torch
from torch import nn
import networkx as nx
from copy import copy # only used in logging.DEBUG mode

LAYER_KEY = 'layers'
NAME_KEY = 'name'
ANCHOR_KEY = 'anchor_layer'
LOSS_KEY = 'loss'
LOSS_REG_KEY = 'loss_weight'
AUTO_WEIGHT_KEY = 'auto'
WEIGHT_INIT_KEY = 'loss_init_val'

MISSING_WEIGHT_MSG = "Expect {0} for task {1} but none provided."

class MTLModel(nn.Module):
    """
    A torch.nn.Module built from a set of shared and task specific layers

    Attributes
    ----------
    g : networkx.Graph
        The meta-computation graph

    task_layers : list
        A list which holds the layers for which to build the computation graph

    output_tasks : list
        A list which holds the tasks for which the output should be returned

    layer_names : list
        A list of the names of each layer

    losses : dict
        A dictionary which maps the name of a layer to its loss function

    loss_weights : dict
        A dictionary which maps the name of a layer to the weight of its loss
        function
    """

    def __init__(self, task_layers, output_tasks):
        super(MTLModel, self).__init__()
        self.task_layers = task_layers
        self.output_tasks = output_tasks
        self.layer_names = [t[NAME_KEY] for t in task_layers]

        self._initialize_graph()

        self._initialize_losses()
        self._initialize_loss_weights()

    def _initialize_losses(self):
        self.losses = {task[NAME_KEY]: task[LOSS_KEY]\
                       for task in self.task_layers if LOSS_KEY in task.keys()}

    def _initialize_loss_weights(self):
        self.loss_weights = {}
        for task in self.task_layers:
            self._set_loss_weight(task)

    def _set_loss_weight(self, task):
        task_name = task[NAME_KEY]
        if LOSS_REG_KEY in task.keys():
            if task[LOSS_REG_KEY] == AUTO_WEIGHT_KEY:
                assert WEIGHT_INIT_KEY in task.keys(),\
                        MISSING_WEIGHT_MSG.format(WEIGHT_INIT_KEY, task_name)
                loss_weight = task[WEIGHT_INIT_KEY]
                loss_name = f'{task_name}_loss'
                loss_weight = torch.nn.Parameter(torch.full((1,),
                                                            loss_weight))
                setattr(self, loss_name, loss_weight)
                self.loss_weights[task_name] = getattr(self, loss_name)
            else:
                self.loss_weights[task_name] = task[LOSS_REG_KEY]
    
    def _initialize_graph(self):
        self.g = nx.DiGraph()
        self.g.add_node('root')
        self._build_graph()

    def _bfs_forward(self, start_node):
        ''' Here we iteratore through the graph in a BFS-fashion starting from
        `start_node`, typically this is the `root` node. This node is skipped
        and we pass the input data and resulting outputs from all layers foward.
        '''
        visited = {node: False for node in self.layer_names}

        # First node is visited
        queue = [start_node]
        visited[start_node] = True

        while queue:
            node = queue.pop(0)
            if node != start_node:
                input_nodes = self.g.predecessors(node)
                if logging.getLogger().level == logging.DEBUG:
                    l = copy(input_nodes)
                    print(f"Feeding output from {list(l)} into {node}")
                cur_layer = getattr(self, node)

                # Get the output from the layers that serve as input
                output_pre_layers = []
                output_complete = True
                for n in input_nodes:
                    # If an output is not ready yet, because that node has not
                    # been computed, we put the current node back into the queue
                    if n not in self.outputs.keys():
                        if logging.getLogger().level == logging.DEBUG:
                            print(f"No output for layer {n} yet")
                        output_complete = False
                        break
                    else:
                        output_pre_layers.append(self.outputs[n])

                if not output_complete:
                    if logging.getLogger().level == logging.DEBUG:
                        print(f"Putting {node} back into the queue.")
                    queue.append(node)
                else:
                    cur_output = cur_layer(*output_pre_layers)
                    self.outputs[node] = cur_output

            for i in self.g.successors(node):
                if visited[i] == False:
                    queue.append(i)
                    visited[i] = True

        losses, loss_weights = self._get_losses()
        return [self.outputs[t] for t in self.output_tasks], losses, loss_weights

    def _get_losses(self):
        losses = []
        loss_weights = []
        for t in self.output_tasks:
            losses.append(self.losses.get(t))
            loss_weights.append(self.loss_weights.get(t))
        return losses, loss_weights

    def _build_graph(self):
        for layer in self.task_layers:
            self._add_layer(layer)
            self._add_to_graph(layer)

    def _add_to_graph(self, layer):
        layer_name = layer[NAME_KEY]
        self._add_node(layer_name)

        if 'anchor_layer' not in layer.keys():
            # If there is no anchor layer, we expect it to be a layer which
            # receives data inputs and is hence connected to the root node
            self.g.add_edge('root', layer_name)
        else:
            anchor_layer = layer[ANCHOR_KEY]
            if isinstance(anchor_layer, list):
                for a_l_name in anchor_layer:
                    self._add_node(a_l_name)
                    self.g.add_edge(a_l_name, layer_name)
            else:
                self._add_node(anchor_layer)
                self.g.add_edge(anchor_layer, layer_name)

    def _add_node(self, layer):
        if isinstance(layer, str):
            layer_name = layer
            self.g.add_node(layer_name)
        else:
            layer_name = layer[NAME_KEY]
            self.g.add_node(layer_name)
            if 'anchor_layer' not in layer.keys():
                self.g.add_edge('root', layer_name)
    
    def _add_layer(self, layer):
        layer_modules = layer[LAYER_KEY]
        layer_name_main = layer[NAME_KEY]
        setattr(self, layer_name_main, layer_modules)

    def forward(self, input):
        self.outputs = {'root': input}
        return self._bfs_forward('root')




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['torchmtl']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.3.2,<4.0.0',
 'networkx>=2.5,<3.0',
 'scipy>=1.5.3,<2.0.0',
 'torch>=1.6.0,<2.0.0']

setup_kwargs = {
    'name': 'torchmtl',
    'version': '0.1.8',
    'description': 'A lightweight module for Multi-Task Learning in pytorch',
    'long_description': '![torchMTL Logo](https://github.com/chrisby/torchMTL/blob/main/images/torchmtl_logo.png "torchMTL Logo")    \nA lightweight module for Multi-Task Learning in pytorch.\n\n`torchmtl` tries to help you composing modular multi-task architectures with minimal effort. All you need is a list of dictionaries in which you define your layers and how they build on each other. From this, `torchmtl` constructs a meta-computation graph which is executed in each forward pass of the created `MTLModel`. To combine outputs from multiple layers, simple [wrapper functions](https://github.com/chrisby/torchMTL/blob/main/torchmtl/wrapping_layers.py) are provided.\n\n### Installation\n`torchmtl` can be installed via `pip`:\n```\npip install torchmtl\n```\n\n### Quickstart (or find examples [here](https://github.com/chrisby/torchMTL/tree/main/examples))\nAssume you want to train a network on three tasks as shown below.  \n![example](https://github.com/chrisby/torchMTL/blob/main/images/example.png "example")  \n\nTo construct such an architecture with `torchmtl`, you simply have to define the following list\n\n```python\ntasks = [\n        {\n            \'name\': "Embed1",\n            \'layers\': Sequential(*[Linear(16, 32), Linear(32, 8)]),\n            # No anchor_layer means this layer receives input directly\n        },    \n        {\n            \'name\': "Embed2",\n            \'layers\': Sequential(*[Linear(16, 32), Linear(32, 8)]),\n            # No anchor_layer means this layer receives input directly\n        },\n        {\n            \'name\': "CatTask",\n            \'layers\': Concat(dim=1),\n            \'loss_weight\': 1.0,\n            \'anchor_layer\': [\'Embed1\', \'Embed2\']\n        },\n        {\n            \'name\': "Task1",\n            \'layers\': Sequential(*[Linear(8, 32), Linear(32, 1)]),\n            \'loss\': MSELoss(),\n            \'loss_weight\': 1.0,\n            \'anchor_layer\': \'Embed1\'            \n        },\n        {\n            \'name\': "Task2",\n            \'layers\': Sequential(*[Linear(8, 64), Linear(64, 1)]),\n            \'loss\': BCEWithLogitsLoss(),\n            \'loss_weight\': 1.0,\n            \'anchor_layer\': \'Embed2\'            \n        }, \n        {\n            \'name\': "FNN",\n            \'layers\': Sequential(*[Linear(16, 32), Linear(32, 32)]),\n            \'anchor_layer\': \'CatTask\'\n        },\n        {\n            \'name\': "Task3",\n            \'layers\': Sequential(*[Linear(32, 16), Linear(16, 1)]),\n            \'anchor_layer\': \'FNN\',\n            \'loss\': MSELoss(),\n            \'loss_weight\': \'auto\',\n            \'loss_init_val\': 1.0\n        }\n    ]\n```\n\nYou can build your final model with the following lines in which you specify from which layers you would like to receive the output.\n```python\nfrom torchmtl import MTLModel\nmodel = MTLModel(tasks, output_tasks=[\'Task1\', \'Task2\', \'Task3\'])\n```\n\nThis constructs a **meta-computation graph** which is executed in each forward pass of your `model`. You can verify whether the graph was properly built by plotting it using the `networkx` library:\n```python\nimport networkx as nx\npos = nx.planar_layout(model.g)\nnx.draw(model.g, pos, font_size=14, node_color="y", node_size=450, with_labels=True)\n```\n![graph example](https://github.com/chrisby/torchMTL/blob/main/images/torchmtl_graph.png "graph example")  \n\n#### The training loop\nYou can now enter the typical `pytorch` training loop and you will have access to everything you need to update your model:\n```python\nfor X, y in data_loader:\n    optimizer.zero_grad()\n\n    # Our model will return a list of predictions (from the layers specified in `output_tasks`),\n    # loss functions, and regularization parameters (as defined in the tasks variable)\n    y_hat, l_funcs, l_weights = model(X)\n    \n    loss = 0\n    # We can now iterate over the tasks and accumulate the losses\n    for i in range(len(y_hat)):\n        loss += l_weights[i] * l_funcs[i](y_hat[i], y[i])\n    \n    loss.backward()\n    optimizer.step()\n\n```\n\n### Details on the layer definition\nThere are 6 keys that can be specified (`name` and `layers` **must** always be present):  \n\n**`layers`**  \nBasically takes any `nn.Module` that you can think of. You can plug in a transformer or just a handful of fully connected layers.  \n\n**`anchor_layer`**  \nThis defines from which other layer this layer receives its input. Take care that the respective dimensions match.  \n\n**`loss`**  \nThe loss function you want to compute on the output of this layer (`l_funcs`). Can be set to `None` or omitted altogether when only access to the layer\'s output is needed.   \n\n**`loss_weight`**  \nThe scalar with which you want to regularize the respective loss (`l_weights`). If set to `\'auto\'`, a `nn.Parameter` is returned which will be updated through backpropagation. Can be set to `None` or omitted altogether when only access to the layer\'s output is needed.  \n\n**`loss_init_val`**  \nOnly needed if `loss_weight: \'auto\'`. The initialization value of the `loss_weight` parameter.\n\n### Wrapping functions\nNodes of the **meta-computation graph** don\'t have to be pytorch Modules. They can be *concatenation* functions or indexing functions that return a certain element of the input. If your `X` consists of two types of input data `X=[X_1, X_2]`, you can use the `SimpleSelect` layer to select the `X_1` by setting  \n```python\nfrom torchmtl.wrapping_layers import SimpleSelect\n{ ...,\n  \'layers\' = SimpleSelect(selection_axis=0),\n  ...\n}\n```\nIt should be trivial to write your own wrapping layers, but I try to provide useful ones with this library. If you have any layers in mind but no time to implement them, feel free to [open an issue](https://github.com/chrisby/torchMTL/issues).\n\nLogo credits and license: I reused and remixed (moved the dot and rotated the resulting logo a couple times) the pytorch logo from [here](https://github.com/pytorch/pytorch/blob/master/docs/source/_static/img/pytorch-logo-dark.png) (accessed through [wikimedia commons](https://commons.wikimedia.org/wiki/File:Pytorch_logo.png)) which can be used under the [Attribution-ShareAlike 4.0 International](https://creativecommons.org/licenses/by-sa/4.0/deed.en) license. Hence, this logo falls under the same license. \n',
    'author': 'Christian Bock',
    'author_email': 'christian.bock@bsse.ethz.ch',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
